import typing
from datetime import datetime
from enum import IntEnum
from uuid import UUID

import msgspec
import typing_extensions as typing_ext

from elytra.core import CamelBaseModel, ParsableCamelModel, add_decoder

__all__ = (
    "ClubUserPresence",
    "ClubDeeplinkMetadata",
    "ClubDeeplinks",
    "ClubPresence",
    "ClubType",
    "ProfileMetadata",
    "Profile",
    "TitleDeeplinkMetadata",
    "TitleDeeplinks",
    "Club",
    "ClubResponse",
)


def _camel_to_const_snake(s: str) -> str:
    return "".join([f"_{c}" if c.isupper() else c.upper() for c in s]).lstrip("_")


class ClubUserPresence(IntEnum):
    UNKNOWN = -1
    NOT_IN_CLUB = 0
    IN_CLUB = 1
    CHAT = 2
    FEED = 3
    ROSTER = 4
    PLAY = 5
    IN_GAME = 6

    @classmethod
    def from_xbox_api(cls, value: str) -> typing_ext.Self:
        try:
            return cls[_camel_to_const_snake(value)]
        except KeyError:
            # it's not like i forgot a value, it's just that some are
            # literally not documented
            return cls.UNKNOWN


class ClubDeeplinkMetadata(CamelBaseModel):
    page_name: str
    uri: str


class ClubDeeplinks(CamelBaseModel):
    xbox: list[ClubDeeplinkMetadata]
    pc: list[ClubDeeplinkMetadata]


class ClubPresence(CamelBaseModel):
    xuid: str
    last_seen_timestamp: datetime
    _last_seen_state: str = msgspec.field(name="lastSeenState")

    @property
    def last_seen_state(self) -> ClubUserPresence:
        return ClubUserPresence.from_xbox_api(self._last_seen_state)


class ClubType(CamelBaseModel):
    type: str
    genre: str
    localized_title_family_name: str
    title_family_id: UUID


class ProfileMetadata(CamelBaseModel):
    can_viewer_change_setting: bool
    value: typing.Optional[typing.Any] = None
    allowed_values: typing.Optional[typing.Any] = None


class Profile(CamelBaseModel):
    description: ProfileMetadata
    rules: ProfileMetadata
    name: ProfileMetadata
    short_name: ProfileMetadata
    is_searchable: ProfileMetadata
    is_recommendable: ProfileMetadata
    request_to_join_enabled: ProfileMetadata
    open_join_enabled: ProfileMetadata
    leave_enabled: ProfileMetadata
    transfer_ownership_enabled: ProfileMetadata
    mature_content_enabled: ProfileMetadata
    watch_club_titles_only: ProfileMetadata
    display_image_url: ProfileMetadata
    background_image_url: ProfileMetadata
    preferred_locale: ProfileMetadata
    tags: ProfileMetadata
    associated_titles: ProfileMetadata
    primary_color: ProfileMetadata
    secondary_color: ProfileMetadata
    tertiary_color: ProfileMetadata


class TitleDeeplinkMetadata(CamelBaseModel):
    title_id: str
    uri: str = msgspec.field(name="Uri")


class TitleDeeplinks(CamelBaseModel):
    xbox: list[TitleDeeplinkMetadata]
    pc: list[TitleDeeplinkMetadata]
    android: list[TitleDeeplinkMetadata]
    ios: list[TitleDeeplinkMetadata] = msgspec.field(
        name="iOS"
    )  # this makes more sense at least


class Club(CamelBaseModel):
    id: str
    club_type: ClubType
    creation_date_utc: datetime
    glyph_image_url: str
    banner_image_url: str
    followers_count: int
    members_count: int
    moderators_count: int
    recommended_count: int
    requested_to_join_count: int
    club_presence_count: int
    club_presence_today_count: int
    club_presence_in_game_count: int
    club_presence: list[ClubPresence]
    state: str
    report_count: int
    reported_items_count: int
    max_members_per_club: int
    max_members_in_game: int
    owner_xuid: str
    founder_xuid: str
    title_deeplinks: TitleDeeplinks
    profile: Profile
    club_deeplinks: ClubDeeplinks
    suspended_until_utc: typing.Optional[typing.Any] = None
    roster: typing.Optional[typing.Any] = None
    target_roles: typing.Optional[typing.Any] = None
    recommendation: typing.Optional[typing.Any] = None
    settings: typing.Optional[typing.Any] = None
    short_name: typing.Optional[typing.Any] = None


@add_decoder
class ClubResponse(ParsableCamelModel):
    clubs: list[Club]
    search_facet_results: typing.Optional[typing.Any] = None
    recommendation_counts: typing.Optional[typing.Any] = None
    club_deeplinks: typing.Optional[typing.Any] = None



import typing

import aiohttp


class HandlerProtocol(typing.Protocol):
    async def request(
        self,
        method: str,
        url: str,
        json: typing.Any = None,
        data: typing.Optional[dict] = None,
        params: typing.Optional[dict] = None,
        headers: typing.Optional[dict] = None,
        *,
        force_refresh: bool = False,
        dont_handle_ratelimit: bool = False,
        **kwargs: typing.Any,
    ) -> aiohttp.ClientResponse:  # type: ignore
        pass

    async def get(
        self,
        url: str,
        json: typing.Any = None,
        data: typing.Optional[dict] = None,
        params: typing.Optional[dict] = None,
        headers: typing.Optional[dict] = None,
        **kwargs: typing.Any,
    ) -> aiohttp.ClientResponse:  # type: ignore
        pass

    async def post(
        self,
        url: str,
        json: typing.Any = None,
        data: typing.Optional[dict] = None,
        params: typing.Optional[dict] = None,
        headers: typing.Optional[dict] = None,
        **kwargs: typing.Any,
    ) -> aiohttp.ClientResponse:  # type: ignore
        pass

    async def put(
        self,
        url: str,
        json: typing.Any = None,
        data: typing.Optional[dict] = None,
        params: typing.Optional[dict] = None,
        headers: typing.Optional[dict] = None,
        **kwargs: typing.Any,
    ) -> aiohttp.ClientResponse:  # type: ignore
        pass

    async def delete(
        self,
        url: str,
        json: typing.Any = None,
        data: typing.Optional[dict] = None,
        params: typing.Optional[dict] = None,
        headers: typing.Optional[dict] = None,
        **kwargs: typing.Any,
    ) -> aiohttp.ClientResponse:  # type: ignore
        pass



import typing
from datetime import datetime

from elytra.core import CamelBaseModel, ParsableCamelModel, PascalBaseModel, add_decoder

__all__ = (
    "PeopleSummaryResponse",
    "Suggestion",
    "Recommendation",
    "MultiplayerSummary",
    "RecentPlayer",
    "Follower",
    "PreferredColor",
    "PresenceDetail",
    "TitlePresence",
    "Detail",
    "SocialManager",
    "Avatar",
    "LinkedAccount",
    "Person",
    "RecommendationSummary",
    "FriendFinderState",
    "PeopleHubResponse",
)


class PeopleSummaryResponse(CamelBaseModel):
    target_following_count: int
    target_follower_count: int
    is_caller_following_target: bool
    is_target_following_caller: bool
    has_caller_marked_target_as_favorite: bool
    has_caller_marked_target_as_identity_shared: bool
    legacy_friend_status: str
    available_people_slots: typing.Optional[int] = None
    recent_change_count: typing.Optional[int] = None
    watermark: typing.Optional[str] = None


# microsoft is a great and consistent company
class Suggestion(PascalBaseModel):
    priority: int
    type: typing.Optional[str] = None
    reasons: typing.Optional[str] = None
    title_id: typing.Optional[str] = None


class Recommendation(PascalBaseModel):
    type: str
    reasons: list[str]


class MultiplayerSummary(PascalBaseModel):
    in_multiplayer_session: int
    in_party: int


class RecentPlayer(CamelBaseModel):
    titles: list[str]
    text: typing.Optional[str] = None


class Follower(CamelBaseModel):
    text: typing.Optional[str] = None
    followed_date_time: typing.Optional[datetime] = None


class PreferredColor(CamelBaseModel):
    primary_color: typing.Optional[str] = None
    secondary_color: typing.Optional[str] = None
    tertiary_color: typing.Optional[str] = None


class PresenceDetail(PascalBaseModel):
    is_broadcasting: bool
    device: str
    presence_text: str
    state: str
    title_id: str
    is_primary: bool
    is_game: bool
    title_type: typing.Optional[str] = None
    rich_presence_text: typing.Optional[str] = None


class TitlePresence(PascalBaseModel):
    is_currently_playing: bool
    presence_text: typing.Optional[str] = None
    title_name: typing.Optional[str] = None
    title_id: typing.Optional[str] = None


class Detail(CamelBaseModel):
    account_tier: str
    is_verified: bool
    watermarks: list[str]
    blocked: bool
    mute: bool
    follower_count: int
    following_count: int
    has_game_pass: bool
    bio: typing.Optional[str] = None
    location: typing.Optional[str] = None
    tenure: typing.Optional[str] = None


class SocialManager(CamelBaseModel):
    title_ids: list[str]
    pages: list[str]


class Avatar(CamelBaseModel):
    update_time_offset: typing.Optional[datetime] = None
    spritesheet_metadata: typing.Optional[typing.Any] = None


class LinkedAccount(CamelBaseModel):
    network_name: str
    show_on_profile: bool
    is_family_friendly: bool
    display_name: typing.Optional[str] = None
    deeplink: typing.Optional[str] = None


class Person(CamelBaseModel):
    xuid: str
    is_favorite: bool
    is_following_caller: bool
    is_followed_by_caller: bool
    is_identity_shared: bool
    real_name: str
    display_pic_raw: str
    show_user_as_avatar: str
    gamertag: str
    gamer_score: str
    modern_gamertag: str
    modern_gamertag_suffix: str
    unique_modern_gamertag: str
    xbox_one_rep: str
    presence_state: str
    presence_text: str
    color_theme: str
    preferred_flag: str
    is_broadcasting: bool
    preferred_platforms: list[str]
    is_quarantined: bool
    is_xbox360_gamerpic: bool
    presence_devices: typing.Optional[typing.Any] = None
    is_cloaked: typing.Optional[bool] = None
    added_date_time_utc: typing.Optional[datetime] = None
    display_name: typing.Optional[str] = None
    suggestion: typing.Optional[Suggestion] = None
    recommendation: typing.Optional[Recommendation] = None
    search: typing.Optional[typing.Any] = None
    title_history: typing.Optional[typing.Any] = None
    multiplayer_summary: typing.Optional[MultiplayerSummary] = None
    recent_player: typing.Optional[RecentPlayer] = None
    follower: typing.Optional[Follower] = None
    preferred_color: typing.Optional[PreferredColor] = None
    presence_details: typing.Optional[list[PresenceDetail]] = None
    title_presence: typing.Optional[TitlePresence] = None
    title_summaries: typing.Optional[typing.Any] = None
    presence_title_ids: typing.Optional[list[str]] = None
    detail: typing.Optional[Detail] = None
    community_manager_titles: typing.Optional[typing.Any] = None
    social_manager: typing.Optional[SocialManager] = None
    broadcast: typing.Optional[list[typing.Any]] = None
    tournament_summary: typing.Optional[typing.Any] = None
    avatar: typing.Optional[Avatar] = None
    linked_accounts: typing.Optional[list[LinkedAccount]] = None
    last_seen_date_time_utc: typing.Optional[datetime] = None


class RecommendationSummary(CamelBaseModel):
    friend_of_friend: int
    facebook_friend: int
    phone_contact: int
    follower: int
    VIP: int
    steam_friend: int
    promote_suggestions: bool


class FriendFinderState(CamelBaseModel):
    facebook_opt_in_status: str
    facebook_token_status: str
    phone_opt_in_status: str
    phone_token_status: str
    steam_opt_in_status: str
    steam_token_status: str
    discord_opt_in_status: str
    discord_token_status: str
    instagram_opt_in_status: str
    instagram_token_status: str
    mixer_opt_in_status: str
    mixer_token_status: str
    reddit_opt_in_status: str
    reddit_token_status: str
    twitch_opt_in_status: str
    twitch_token_status: str
    twitter_opt_in_status: str
    twitter_token_status: str
    you_tube_opt_in_status: str
    you_tube_token_status: str


@add_decoder
class PeopleHubResponse(ParsableCamelModel):
    people: list[Person]
    recommendation_summary: typing.Optional[RecommendationSummary] = None
    friend_finder_state: typing.Optional[FriendFinderState] = None
    account_link_details: typing.Optional[list[LinkedAccount]] = None



from elytra.const import BEDROCK_REALMS_API_URL, MC_VERSION
from elytra.core import BaseMicrosoftAPI

from .models import *

__all__ = (
    "Permission",
    "State",
    "WorldType",
    "FullRealm",
    "MultiRealmResponse",
    "Player",
    "PartialRealm",
    "ActivityListResponse",
    "PendingInvite",
    "PendingInviteResponse",
    "BedrockRealmsAPI",
)


class BedrockRealmsAPI(BaseMicrosoftAPI):
    RELYING_PATH: str = BEDROCK_REALMS_API_URL
    BASE_URL: str = BEDROCK_REALMS_API_URL

    @property
    def base_headers(self) -> dict[str, str]:
        return {
            "Authorization": self.auth_mgr.xsts_token.authorization_header_value,
            "Client-Version": MC_VERSION,
            "User-Agent": "MCPE/UWP",
        }

    async def join_realm_from_code(self, code: str) -> FullRealm:
        return await FullRealm.from_response(
            await self.post(f"invites/v1/link/accept/{code}")
        )

    async def fetch_realms(self) -> MultiRealmResponse:
        return await MultiRealmResponse.from_response(await self.get("worlds"))

    async def invite_player(
        self, realm_id: str | int, player_xuid: str | int
    ) -> FullRealm:
        return await FullRealm.from_response(
            await self.put(
                f"invites/{realm_id}/invite/update",
                json={"invites": {str(player_xuid): "ADD"}},
            )
        )

    async def fetch_pending_invite_count(self) -> int:
        resp = await self.get("invites/count/pending")
        return int(await resp.text())

    async def fetch_pending_invites(self) -> PendingInviteResponse:
        return await PendingInviteResponse.from_response(
            await self.get("invites/pending")
        )

    async def accept_invite(self, invitation_id: str) -> None:
        await self.put(f"invites/accept/{invitation_id}")

    async def reject_invite(self, invitation_id: str) -> None:
        await self.put(f"invites/reject/{invitation_id}")

    async def fetch_activities(self) -> ActivityListResponse:
        return await ActivityListResponse.from_response(
            await self.get("activities/live/players")
        )

    async def leave_realm(self, realm_id: int | str) -> None:
        await self.delete(f"invites/{realm_id}")



import asyncio
import typing

import aiohttp
import aiohttp_retry


class RTAHandler:
    session: aiohttp_retry.RetryClient

    async def _rta_start(
        self,
        url: str,
        dispatch_handler: typing.Callable[
            [aiohttp.WSMessage], typing.Awaitable[typing.Any]
        ],
    ) -> None:
        async with self.session._client.ws_connect(
            "wss://rta.xboxlive.com/connect"
        ) as ws:
            to_send = f'[1,1,"{url}"]'
            await ws.send_str(to_send)

            async for msg in ws:
                await dispatch_handler(msg)

    async def rta(
        self,
        url: str,
        dispatch_handler: typing.Callable[
            [aiohttp.WSMessage], typing.Awaitable[typing.Any]
        ],
    ) -> None:
        asyncio.create_task(self._rta_start(url, dispatch_handler))



from elytra.core import CamelBaseModel, ParsableCamelModel, add_decoder

__all__ = ("Setting", "ProfileUser", "ProfileResponse")


class Setting(CamelBaseModel):
    id: str
    value: str


class ProfileUser(CamelBaseModel):
    id: str
    host_id: str
    settings: list[Setting]
    is_sponsored_user: bool


@add_decoder
class ProfileResponse(ParsableCamelModel):
    profile_users: list[ProfileUser]



__all__ = ("XBOX_API_RELYING_PARTY", "BEDROCK_REALMS_API_URL", "MC_VERSION")

XBOX_API_RELYING_PARTY = "http://xboxlive.com"
BEDROCK_REALMS_API_URL = "https://pocket.realms.minecraft.net/"
MC_VERSION = "1.20.0"  # this can be a few versions behind



import argparse
import asyncio
import os
from urllib.parse import urlencode

import aiohttp
import aiohttp.web
import aiohttp_retry
import msgspec

from elytra import AuthenticationManager

code: str | None = None


def generate_authorization_url(client_id: str, redirect_uri: str) -> str:
    """Generate Windows Live Authorization URL."""
    query_params = {
        "client_id": client_id,
        "response_type": "code",
        "approval_prompt": "auto",
        "scope": "Xboxlive.signin Xboxlive.offline_access",
        "redirect_uri": redirect_uri,
    }

    return f"https://login.live.com/oauth20_authorize.srf?{urlencode(query_params)}"


routes = aiohttp.web.RouteTableDef()


@routes.get("/auth/callback")
async def parse_response(req: aiohttp.web.Request) -> aiohttp.web.Response:
    global code

    try:
        if req.query.get("error"):
            return aiohttp.web.Response(
                text=f"Error: {req.query.get('error_description')}", status=400
            )

        if gotten_code := req.query.get("code"):
            code = gotten_code
            return aiohttp.web.Response(
                text="Success! You may now close this tab and go back to the terminal."
            )

        return aiohttp.web.Response(text="Error: No code was provided.", status=400)
    finally:
        app["event"].set()


app = aiohttp.web.Application()
app.add_routes(routes)


async def async_main() -> None:
    global code

    parser = argparse.ArgumentParser(description="Authenticate into Microsoft's APIs.")
    parser.add_argument(
        "--file",
        "-f",
        default="oauth.json",
        help="Filepath to put output of program. Default: 'oauth.json'.",
    )
    parser.add_argument(
        "--client-id",
        "-cid",
        default=os.environ.get("CLIENT_ID"),
        help="OAuth2 client ID.",
    )
    parser.add_argument(
        "--client-secret",
        default=os.environ.get("CLIENT_SECRET"),
        help="OAuth2 client secret.",
    )
    args = parser.parse_args()

    if not (args.client_id and args.client_secret):
        raise ValueError(
            "You must provide a client ID and secret either via the CLI or environment"
            " variables."
        )

    app["event"] = asyncio.Event()

    redirect_uri = "http://localhost:8080/auth/callback"

    runner = aiohttp.web.AppRunner(app)
    await runner.setup()
    site = aiohttp.web.TCPSite(runner, "localhost", 8080)
    await site.start()

    print(  # noqa: T201
        f"Please visit {generate_authorization_url(args.client_id, redirect_uri)} to"
        " authenticate."
    )

    await app["event"].wait()
    await runner.cleanup()

    if code:
        auth_mgr = AuthenticationManager(
            aiohttp_retry.RetryClient(),
            args.client_id,
            args.client_secret,
            "http://xboxlive.com",
        )
        oauth = await auth_mgr.request_oauth_token(code, redirect_uri)
        with open(args.file, mode="wb") as f:
            f.write(msgspec.json.encode(oauth))

        await auth_mgr.close()
        print("Authentication successful!")  # noqa: T201
    else:
        print("Authentication failed.")  # noqa: T201


def main() -> None:
    asyncio.run(async_main())


if __name__ == "__main__":
    main()



import typing

from elytra.protocols import HandlerProtocol

from .models import *

__all__ = ("ProfileHandler", "ProfileUser", "ProfileResponse", "Setting")


class ProfileHandler(HandlerProtocol):
    async def fetch_profiles(
        self, xuid_list: list[str] | list[int], **kwargs: typing.Any
    ) -> ProfileResponse:
        URL = "https://profile.xboxlive.com/users/batch/profile/settings"
        HEADERS = {"x-xbl-contract-version": "3"}

        post_data = {
            "settings": ["Gamertag"],
            "userIds": xuid_list,
        }
        return await ProfileResponse.from_response(
            await self.post(URL, json=post_data, headers=HEADERS, **kwargs)
        )

    async def fetch_profile_by_xuid(
        self, target_xuid: str | int, **kwargs: typing.Any
    ) -> ProfileResponse:
        HEADERS = {"x-xbl-contract-version": "3"}
        PARAMS = {"settings": "Gamertag"}
        URL = f"https://profile.xboxlive.com/users/xuid({target_xuid})/profile/settings"
        return await ProfileResponse.from_response(
            await self.get(URL, params=PARAMS, headers=HEADERS, **kwargs)
        )

    async def fetch_profile_by_gamertag(
        self, gamertag: str, **kwargs: typing.Any
    ) -> ProfileResponse:
        url = f"https://profile.xboxlive.com/users/gt({gamertag})/profile/settings"
        HEADERS = {"x-xbl-contract-version": "3"}
        PARAMS = {"settings": "Gamertag"}

        return await ProfileResponse.from_response(
            await self.get(url, params=PARAMS, headers=HEADERS, **kwargs)
        )



import argparse
import asyncio
import os

import aiohttp
import msgspec


async def async_main() -> None:
    parser = argparse.ArgumentParser(description="Authenticate into Microsoft's APIs.")
    parser.add_argument(
        "--file",
        "-f",
        default="oauth.json",
        help="Filepath to put output of program. Default: 'oauth.json'.",
    )
    parser.add_argument(
        "--client-id",
        "-cid",
        default=os.environ.get("CLIENT_ID"),
        help="OAuth2 client ID.",
    )
    args = parser.parse_args()

    if not args.client_id:
        raise ValueError(
            "You must provide a client ID either via the CLI or environment variables."
        )

    async with aiohttp.ClientSession() as session:
        async with session.post(
            "https://login.microsoftonline.com/consumers/oauth2/v2.0/devicecode",
            data={
                "client_id": args.client_id,
                "scope": "Xboxlive.signin Xboxlive.offline_access",
            },
        ) as r:
            data = await r.json()
            if not r.ok:
                print(data)  # noqa: T201
                return

        success_response: dict | None = None

        print(data["message"])  # noqa: T201

        async with asyncio.Timeout(data["expires_in"]):
            while True:
                await asyncio.sleep(data["interval"])
                async with session.post(
                    "https://login.microsoftonline.com/consumers/oauth2/v2.0/token",
                    data={
                        "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
                        "client_id": args.client_id,
                        "device_code": data["device_code"],
                    },
                ) as r:
                    resp_json = await r.json()
                    if error := resp_json.get("error"):
                        if error in {
                            "authorization_declined",
                            "expired_token",
                            "bad_verification_code",
                        }:
                            break
                    else:
                        success_response = resp_json
                        break

        if not success_response:
            print("Authentication failed.")  # noqa: T201
            return

        with open(args.file, "wb") as f:
            f.write(msgspec.json.encode(success_response))

        print("Authentication successful!")  # noqa: T201


def main() -> None:
    asyncio.run(async_main())


if __name__ == "__main__":
    main()
